/*
 * Basic template code for getting the gyro x, y, z axis data
 * from the MPU6050
 * 
 * Connections:
 * MPU6050    Arduino  or   ESP8266-12
 *   SDA        A4          GPIO 04
 *   SCL        A5          GPIO 05
 *   INT        2           GPIO 15 (optional)
 *   VCC        5V          3.3V
 *   GND        GND         GND
 * 
 * This code is based on Jeff Rowberg's <jeff@rowberg.net> MPU6050 library
 * https://github.com/jrowberg/i2cdevlib/tree/master/Arduino/MPU6050
 * 
 * Copyright (c) 2020 Enoch Hwang
 */


//////////////////////////////////////////////////////////////
//// MPU6050 stuff
#include "MPU6050_6Axis_MotionApps20.h"
MPU6050 mpu;
//#define USE_INTERRUPT     // define this to use interrupts
                            // the INT pin must be connected if using interrupts
#define INTERRUPT_PIN 2     // must use pin 2 on Arduino Uno & most boards
//#define INTERRUPT_PIN 15  // must use pin 15 on ESP8266
volatile bool mpuInterrupted = false;     // indicates whether MPU interrupt pin has gone high

//// Interrupt detection routine
void dmpDataReady() {
    mpuInterrupted = true;
}

//// initialize and calibrate the MPU
//// return true if successful
bool initMPU() {
  uint8_t devStatus;      // return status after each device operation (0 = success, !0 = error)
  // join I2C bus (I2Cdev library doesn't do this automatically)
  #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
      Wire.begin();
      Wire.setClock(400000); // 400kHz I2C clock. Comment this line if having compilation difficulties
  #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
      Fastwire::setup(400, true);
  #endif

  // initialize device
  Serial.println("Initializing I2C devices...");
  mpu.initialize();
  pinMode(INTERRUPT_PIN, INPUT);

  // verify connection
  Serial.println("Testing device connections...");
  Serial.println(mpu.testConnection() ? "MPU6050 connection successful" : "MPU6050 connection failed");

  // load and configure the DMP
  Serial.println("Initializing DMP...");
  devStatus = mpu.dmpInitialize();

  // make sure it worked (returns 0 if so)
  if (devStatus == 0) {
    // Calibration Time: generate offsets and calibrate our MPU6050
    Serial.print("Calibrating DMP...");
    mpu.CalibrateAccel(6);
    mpu.CalibrateGyro(6);
    mpu.PrintActiveOffsets();
    // turn on the DMP, now that it's ready
    Serial.println("Enabling DMP...");
    mpu.setDMPEnabled(true);

    // enable Arduino interrupt detection
    Serial.print("Enabling interrupt detection (Arduino external interrupt #");
    Serial.print(digitalPinToInterrupt(INTERRUPT_PIN));
    Serial.print(" on pin ");
    Serial.print(INTERRUPT_PIN);
    Serial.println(")...");
    attachInterrupt(digitalPinToInterrupt(INTERRUPT_PIN), dmpDataReady, RISING);

    // set our DMP Ready flag so the main loop() function knows it's okay to use it
    Serial.println("DMP ready! Waiting for first interrupt...");
    return true;

  } else {
    // ERROR!
    // 1 = initial memory load failed
    // 2 = DMP configuration updates failed
    // (if it's going to break, usually the code will be 1)
    Serial.print("DMP Initialization failed (code ");
    Serial.print(devStatus);
    Serial.println(")");
    while(1); // stop here if init failed
    return false;
  }
}


//// Check if there's data from the MPU
//// and if there's data then
//// get and return the gyro x, y, z axis data
//// This routine must be called very often
//// Otherwise the FIFO will overflow
void checkMPU(float xyz[3]) {
  // MPU control/status vars
  static uint8_t mpuIntStatus;   // holds actual interrupt status byte from MPU
  static uint16_t fifoCount;     // count of all bytes currently in FIFO
  static uint8_t fifoBuffer[64]; // FIFO storage buffer
  static uint16_t packetSize = mpu.dmpGetFIFOPacketSize();    // expected DMP packet size for later comparison (default is 42 bytes)
  // orientation/motion vars
  static Quaternion q;           // [w, x, y, z]         quaternion container
  static VectorFloat gravity;    // [x, y, z]            gravity vector
  static float ypr[3];           // [yaw, pitch, roll]   yaw/pitch/roll container and gravity vector

#ifdef USE_INTERRUPT
  // check if there's data in the fifo
  if (!mpuInterrupted && fifoCount < packetSize) {
    return; // no data yet, so return
  }
#endif

  // reset interrupt flag and get INT_STATUS byte
  mpuInterrupted = false;
  mpuIntStatus = mpu.getIntStatus();

  // get current FIFO count
  fifoCount = mpu.getFIFOCount();
  
  // check for overflow (this should never happen unless our code is too inefficient)
  if ((mpuIntStatus & _BV(MPU6050_INTERRUPT_FIFO_OFLOW_BIT)) || fifoCount >= 1024) {
      // reset so we can continue cleanly
      mpu.resetFIFO();
      //  fifoCount = mpu.getFIFOCount();  // will be zero after reset no need to ask
      Serial.println("FIFO overflow!");
  } 
  
  // otherwise, check for DMP data ready interrupt (this should happen frequently)
  else if (mpuIntStatus & _BV(MPU6050_INTERRUPT_DMP_INT_BIT)) {
    // wait for correct available data length, should be a VERY short wait
    while (fifoCount < packetSize) fifoCount = mpu.getFIFOCount();
    // read a packet from FIFO
    mpu.getFIFOBytes(fifoBuffer, packetSize);

    // get and return the gyro x, y, z axis data
    mpu.dmpGetQuaternion(&q, fifoBuffer);
    mpu.dmpGetGravity(&gravity, &q);
    mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
    xyz[0] = ypr[2] * 180/M_PI;   // roll(x axis)
    xyz[1] = ypr[1] * 180/M_PI;   // pitch(y axis)
    xyz[2] = ypr[0] * 180/M_PI;   // yaw(z axis)
  }
}


void setup() {
  Serial.begin(115200);
  initMPU();
}

void loop() {
  static float xyz[3];
  checkMPU(xyz);
  Serial.print("x = ");
  Serial.print(xyz[0]);
  Serial.print("\t\ty = ");
  Serial.print(xyz[1]);
  Serial.print("\t\tz = ");
  Serial.print(xyz[2]);
  Serial.println();
}
